#  ___________________________________________________________________________
#
#  Pyomo: Python Optimization Modeling Objects
#  Copyright 2017 National Technology and Engineering Solutions of Sandia, LLC
#  Under the terms of Contract DE-NA0003525 with National Technology and 
#  Engineering Solutions of Sandia, LLC, the U.S. Government retains certain 
#  rights in this software.
#  This software is distributed under the 3-clause BSD License.
#  ___________________________________________________________________________
#
# Pyomo plugin for interacting with NEOS solvers
#

# Static documentation for NEOS solvers
doc = {
'bonmin':   'Heuristic MINLP solver',
'cbc':      'MILP solver',
'conopt':   'Feasible path NLP solver',
'couenne':  'Exact MINLP solver',
'cplex':    'MILP solver',
'filmint':  'Heuristic MINLP solver',
'filter':   'SQP NLP solver',
'ipopt':    'Interior point NLP solver',
'knitro':   'Convex MINLP solver',
'l-bfgs-b': 'Bound-constrained NLP solver',
'lancelot': 'Augmented Lagrangian NLP solver',
'loqo':     'Interior point NLP solver',
'minlp':    'Heuristic MINLP solver',
'minos':    'SLC NLP solver',
'minto':    'MILP solver',
'mosek':    'Interior point NLP solver',
'ooqp':     'Convex QP solver',
'path':     'Nonlinear MCP solver',
'snopt':    'SQP NLP solver',
'raposa':   'A Global Solver for Polynomial Programming Problems',
'lgo':      'Lipschitz-Continuous Global Optimizer'
}

